/*******************************************************************************
 * Copyright (c) 2000, 2004 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.team.internal.ftp.client;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;

import org.eclipse.team.internal.ftp.FTPException;

/**
 * Factory for converting the following LS line to an FTPFDirectoryEntry:
 * 
 * drwxr-xr-x    5 1129     1000         4096 Dec 18 15:44 downloads
 * 
 */
public class LSDirectoryEntryParser extends DirectoryEntryParser {

	protected static final String[] months = new String[] {
			"jan", "feb", "mar", "apr", "may", "jun", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$
			"jul", "aug", "sep", "oct", "nov", "dec" //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$
		};
	
	protected boolean hasDirectorySementics(String line) {
		switch (line.charAt(0)) { // see man page for /bin/ls
			case 'd': // directory
				return true;
			case 'l': // symbolic link, don't know if this is a directory or a file
				return true;
			default:
				return false;
		}
	}
	
	protected boolean hasFileSementics(String line) {
		switch (line.charAt(0)) { // see man page for /bin/ls
			case 'd': // directory
				return false;
			case 'l': // symbolic link, don't know if this is a directory or a file
				return true;
			case '-': // regular file
				return true;
			default:
				return false;
		}
	}
	
	/**
	 * Parses the date into the given calendar and returns the name.
	 * Recognizes the following formats:
	 * <ul>
	 *   <li>MMM dd [hh:mm] [yyyy] filename -> symlink</li>
	 * </ul></p>
	 * @param tokenizer
	 * @param calendar
	 * @return the filename or <code>null</code> if parsing failed
	 */
	protected String parseDateAndName(StringTokenizer tokenizer, Calendar calendar) {
		String token;
		if (! parseMonth(calendar, token = tokenizer.nextToken())) return null;
		if (! parseDay(calendar, token = tokenizer.nextToken())) return null;
		if (! parseTime(calendar, token = tokenizer.nextToken())) {
			calendar.set(Calendar.HOUR_OF_DAY, 0);
			calendar.set(Calendar.MINUTE, 0);
			calendar.set(Calendar.SECOND, 0);
			calendar.set(Calendar.MILLISECOND, 0);
		} else {
			token = tokenizer.nextToken();
		}
		if (! parseYear(calendar, token)) {
			// some servers don't send the year if the date is within the last 6/12 months
			// assume the date is within the last 12 months or so (cannot handle dates in the future)
			Calendar current = GregorianCalendar.getInstance(Constants.DEFAULT_TIMEZONE);
			current.setTime(new Date(System.currentTimeMillis()));
			int currentYear = current.get(Calendar.YEAR);
			int currentDayOfYear = current.get(Calendar.DAY_OF_YEAR);
			int dayOfYear = calendar.get(Calendar.DAY_OF_YEAR);
			if (dayOfYear > currentDayOfYear + 1) { // +1 for leapyears and subtle discrepancies
				calendar.set(Calendar.YEAR, currentYear - 1);
			} else {
				calendar.set(Calendar.YEAR, currentYear);
			}
		} else {
			token = tokenizer.nextToken();
		}
		// read filename
		StringBuffer filename = new StringBuffer(token);
		while (tokenizer.hasMoreTokens()) {
			token = tokenizer.nextToken();
			if (token.equals("->")) break; // followed by the symlink pointer //$NON-NLS-1$
			filename.append(' ');
			filename.append(token);
		}
		return filename.toString();
	}
	
	protected boolean parseMonth(Calendar calendar, String text) {
		for (int i = 0; i < months.length; ++i) {
			if (text.equalsIgnoreCase(months[i])) {
				calendar.set(Calendar.MONTH, i);
				return true;
			}
		}
		return false;
	}
	
	protected boolean parseDay(Calendar calendar, String text) {
		try {
			int day = Integer.parseInt(text, 10);
			if (day > 0 && day < 32) {
				calendar.set(Calendar.DAY_OF_MONTH, day);
				return true;
			}
		} catch (NumberFormatException e) {
		}
		return false;
	}
	
	protected boolean parseYear(Calendar calendar, String text) {
		try {
			int year = Integer.parseInt(text, 10);
			if (year >= 1900 && year <= 2100) {
				calendar.set(Calendar.YEAR, year);
				return true;
			}
		} catch (NumberFormatException e) {
		}
		return false;
	}
	
	protected boolean parseTime(Calendar calendar, String time) {
		try {
			int colonPos = time.indexOf(':');
			if (colonPos != -1) {
				int hour = Integer.parseInt(time.substring(0, colonPos), 10);
				if (hour == 24) hour = 0;
				if (hour >= 0 && hour <= 23) {
					int minute = Integer.parseInt(time.substring(colonPos + 1), 10);
					if (minute >= 0 && minute <= 59) {
						calendar.set(Calendar.HOUR_OF_DAY, hour);
						calendar.set(Calendar.MINUTE, minute);
						calendar.set(Calendar.SECOND, 0);
						calendar.set(Calendar.MILLISECOND, 0);
						return true;
					}
				}
			}
		} catch (NumberFormatException e) {
		}
		return false;
	}
	
	/**
	 * Parses /bin/ls format lines.
	 * <p>
	 * Understands the following formats:
	 * <ul>
	 *   <li>d<permissions> <blocks> <uid> <gid> <size> <date> <dirname></li>
	 *   <li>l<permissions> <blocks> <uid> <gid> <size> <date> <file or dirname> -> <link></li>
	 *   <li>-<permissions> <blocks> <uid> <gid> <size> <date> <filename></li>
	 *   <li>d <permissions> <uid> <size> <date> <filename></li>
	 *   <li>- <permissions> <uid> <size> <date> <filename></li>
	 * </ul>
	 * Recognizes the following date formats:
	 * <ul>
	 *   <li>MMM dd [hh:mm] [yyyy]</li>
	 * </ul></p>
	 * 
	 * @param line the line
	 * @return the file info, or null if the line could not be parsed
	 */
	public FTPDirectoryEntry parseEntry(String line) throws FTPException {
		boolean hasDirectorySemantics = hasDirectorySementics(line);
		boolean hasFileSemantics = hasFileSementics(line);
		if (!hasFileSemantics && !hasDirectorySemantics) {
			return null;
		}
		StringTokenizer tok = new StringTokenizer(line, " \t"); //$NON-NLS-1$
		try {
			
			// read permissions, blocks, uid, gid, size
			String permissions = tok.nextToken();
			if (permissions.length() == 1) {
				// Handle format observed from certain servers
				permissions = tok.nextToken();
				String uid = tok.nextToken();
			} else {
				long blocks = Long.parseLong(tok.nextToken(), 10);
				String uid = tok.nextToken();
				String gid = tok.nextToken();
			}
			long size = Long.parseLong(tok.nextToken(), 10);
			
			// read date (try hard!)
			String token;
			Calendar calendar = GregorianCalendar.getInstance(Constants.DEFAULT_TIMEZONE);
			String filename = parseDateAndName(tok, calendar);
			if (filename == null) {
				return null;
			}
			
			// done!
			return new FTPDirectoryEntry(filename, hasDirectorySemantics, hasFileSemantics,
				size, calendar.getTime());
		} catch (NoSuchElementException e) {
			// expected more tokens
			return null;
		} catch (NumberFormatException e) {
			// don't understand the format so ignore the line
			return null;
		}
	}

	/* (non-Javadoc)
	 * @see org.eclipse.team.internal.ftp.client.DirectoryEntryParser#isReliable()
	 */
	public boolean isReliable() {
		return false;
	}

}
