/*******************************************************************************
 * Copyright (c) 2000, 2004 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.team.internal.target.subscriber;

import java.util.*;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.team.core.TeamException;
import org.eclipse.team.internal.webdav.core.Policy;

/**
 * Operation to upload resource to an Deployment target
 */
public class PutOperation extends TargetOperation {

	private boolean overwriteRemoteChanges;

	public PutOperation(String deploymentPrivderId, IResource[] resources, int depth, boolean overwriteRemoteChanges) {
		super(deploymentPrivderId, resources, depth);
		this.overwriteRemoteChanges = overwriteRemoteChanges;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.team.internal.ftp.deployment.FTPOperation#execute(org.eclipse.core.runtime.IProgressMonitor)
	 */
	protected void execute(TargetDeploymentProvider provider, IResource[] resources, IProgressMonitor monitor) throws TeamException {
		monitor.beginTask(null, 100 * resources.length);
		// Must put all directories first, followed by files
		Set folderSet = new HashSet();
		for (int i = 0; i < resources.length; i++) {
			IResource resource = resources[i];
			if (resource.getType() != IResource.FILE) {
				folderSet.add(resource);
			}
		}
		if (!folderSet.isEmpty()) {
			List folders = new ArrayList(folderSet.size());
			folders.addAll(folderSet);
			Collections.sort(folders, new Comparator() {
				public boolean equals(Object obj) {
					return false;
				}
				public int compare(Object o1, Object o2) {
					IResource r1 = (IResource)o1;
					IResource r2 = (IResource)o2;
					return r1.getFullPath().segmentCount() - r2.getFullPath().segmentCount();
				}
			});
			for (Iterator iter = folders.iterator(); iter.hasNext();) {
				IResource folder = (IResource) iter.next();
				put(provider, folder, getDepth(), Policy.subMonitorFor(monitor, 70));
			}
		}
		
		// Now put all the files
		for (int i = 0; i < resources.length; i++) {
			IResource resource = resources[i];
			if (resource.getType() == IResource.FILE) {
				put(provider, resource, getDepth(), Policy.subMonitorFor(monitor, 70));
			}
		}
		monitor.done();
	}

	public void upload(TargetDeploymentProvider provider, IResource localResource, IProgressMonitor monitor) throws TeamException {
		// Upload only applies to files as folders are created using mkRemoteDirs()
		if (localResource.getType() != IResource.FILE) return;
		
		// Upload the file and set the local and remote base identifiers
		monitor.beginTask(null, 100);
		try {
			IFile localFile = (IFile)localResource;
			putFile(provider, localFile, Policy.subMonitorFor(monitor, 90));
			// The fetching of the timestamp will be delayed.
			// This may result in the lose of remote changes in some circumstances.
			// However, the performance gain is large
			needsRemoteIdentifier(provider, localResource, Policy.subMonitorFor(monitor, 10));
		} finally {
			monitor.done();
		}
	}
	
	protected void needsRemoteIdentifier(TargetDeploymentProvider provider, IResource localResource, IProgressMonitor monitor) throws TeamException {
		getSynchronizer(provider).setBaseBytes(localResource, provider.fetchRemoteSyncBytes(localResource, monitor));
	}
	

	protected void putFile(TargetDeploymentProvider provider, IFile localFile, IProgressMonitor monitor) throws TeamException {
		provider.putFile(localFile, monitor);
	}
	
	protected void createRemoteDirectory(TargetDeploymentProvider provider, IResource resource, IProgressMonitor monitor) throws TeamException {
		provider.createRemoteDirectory(resource, monitor);
		needsRemoteIdentifier(provider, resource, monitor);
	}

	protected final void put(TargetDeploymentProvider provider, IResource localResource, IProgressMonitor progress) throws TeamException {

		progress = Policy.monitorFor(progress);
		progress.beginTask(null, 100);
		try {
			
			if (!localResource.exists()) {
				delete(provider, localResource, Policy.subMonitorFor(progress, 80));
			} else {
				// Copy from the local resource to the repository.
				if (localResource.getType() == IResource.FILE) {
					upload(provider, localResource, Policy.subMonitorFor(progress, 80));
				} else {
					if (!hasRemote(provider, localResource)) {
						createRemoteDirectory(provider, localResource, Policy.subMonitorFor(progress, 80));
					}
				}
			}
		} finally {
			progress.done();
		}
	}

	private void put(TargetDeploymentProvider provider, IResource resource, int depth, IProgressMonitor monitor) throws TeamException {
		IProgressMonitor progress = Policy.monitorFor(monitor);
		try {
			progress.beginTask(null, 100);
			Policy.checkCanceled(progress);
			
			if (resource.exists()) {
				// The resource exists locally so upload it.
				put(provider, resource, Policy.subMonitorFor(progress, 100));
			}
			
			if (depth != IResource.DEPTH_ZERO) {	
				// TODO: May not want to put all memebers as some may be uncontrolled
				IResource[] children = getSynchronizer(provider).members(resource);
				for (int i = 0; i < children.length; i++) {
					IResource localChild = children[i];
					put(provider, localChild, (depth == IResource.DEPTH_INFINITE) ? IResource.DEPTH_INFINITE : IResource.DEPTH_ONE, Policy.subMonitorFor(progress, 100));
				}
			}
			
			if (!resource.exists()) {
				// The resource doesn't exist locally so remove it remotely.
				// This is done last so that children would have been deleted first.
				// This may fail if all children were not removed
				delete(provider, resource, Policy.subMonitorFor(progress, 100));
			}
		} finally {
			progress.done();
		}
	}
	
	private void delete(TargetDeploymentProvider provider, IResource resource, IProgressMonitor progress) throws TeamException {
		provider.deleteRemote(resource, progress);
		getSynchronizer(provider).flush(resource, IResource.DEPTH_INFINITE);
	}
}
