/*******************************************************************************
 * Copyright (c) 2000, 2004 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.team.internal.ui.target;

import java.util.*;

import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceVisitor;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.team.core.TeamException;
import org.eclipse.team.internal.core.target.*;
import org.eclipse.team.internal.ui.Utils;
import org.eclipse.team.internal.ui.actions.TeamAction;

public abstract class TargetAction extends TeamAction {
	/**
	 * Get selected remote target folders
	 */
	protected IRemoteTargetResource[] getSelectedRemoteFolders() {
		ArrayList resources = null;
		if (!selection.isEmpty()) {
			resources = new ArrayList();
			Iterator elements = selection.iterator();
			while (elements.hasNext()) {
				Object next = elements.next();
				IRemoteTargetResource remote = null;
				if (next instanceof RemoteResourceElement) {
					remote = ((RemoteResourceElement)next).getRemoteResource();
				} else if(next instanceof SiteElement) {
					try {
						remote = ((SiteElement)next).getSite().getRemoteResource();
					} catch (TeamException e) {
						Utils.handle(e);
						return new IRemoteTargetResource[0];
					}
				}
				if(remote != null && remote.isContainer()) {
					resources.add(remote);
					continue;
				}
			}
		}
		if (resources != null && !resources.isEmpty()) {
			return (IRemoteTargetResource[])resources.toArray(new IRemoteTargetResource[resources.size()]);
		}
		return new IRemoteTargetResource[0];
	}
	
	/**
	 * Get selected remote target folders
	 */
	protected Site[] getSelectedSites() {
		ArrayList sites = new ArrayList();
		if (!selection.isEmpty()) {
			Iterator elements = selection.iterator();
			while (elements.hasNext()) {
				Object next = elements.next();
				if (next instanceof SiteElement) {
					sites.add(((SiteElement)next).getSite());
				}
			}
		}
		return (Site[])sites.toArray(new Site[sites.size()]);
	}
	
	protected IResource[] findResourcesWithOutgoingChanges(IResource[] resources) throws TeamException, CoreException {
		// Collect the dirty resource		
		final List dirtyResources = new ArrayList();
		for (int i = 0; i < resources.length; i++) {
			IResource resource = resources[i];
			final TargetProvider provider = TargetManager.getProvider(resource.getProject());
			resource.accept(new IResourceVisitor() {
				public boolean visit(IResource resource) throws CoreException {
					if (resource.getType() == IResource.FILE) {
						if (provider.isDirty(resource) || ! provider.hasBase(resource)) {
							dirtyResources.add(resource);
						}
					} else {
						// Check for outgoing folder deletions?
					}
					return true;
				}
			}, IResource.DEPTH_INFINITE, true /* include phantoms */);		
		}
		return (IResource[]) dirtyResources.toArray(new IResource[dirtyResources.size()]);
	}
	
	/**
	 * Convenience method that maps the selected resources to their target providers.
	 * The returned Hashtable has keys which are TargetProviders, and values
	 * which are Lists of IResources that are shared with that provider.
	 * 
	 * @return a hashtable mapping providers to their selected resources
	 */
	protected Hashtable getTargetProviderMapping() throws TeamException {
		return getTargetProviderMapping(getSelectedResources());
	}
	/**
	 * Convenience method that maps the given resources to their target providers.
	 * The returned Hashtable has keys which are TargetProviders, and values
	 * which are Lists of IResources that are shared with that provider.
	 * 
	 * @return a hashtable mapping providers to their resources
	 */
	protected Hashtable getTargetProviderMapping(IResource[] resources) throws TeamException {
		Hashtable result = new Hashtable();
		for (int i = 0; i < resources.length; i++) {
			TargetProvider provider = TargetManager.getProvider(resources[i].getProject());
			List list = (List)result.get(provider);
			if (list == null) {
				list = new ArrayList();
				result.put(provider, list);
			}
			list.add(resources[i]);
		}
		return result;
	}
}
